
#include	<stdlib.h>
#include	<stdio.h>
#include	<math.h>
#include	<string.h>
#include	"matrix.h"
#include	"js_simple.h"

#define	MAXSUPSIZE	50
#define	NUMPROBS	1

static double	prods[NUMPROBS][MAXSUPSIZE][MAXSUPSIZE];
static double	eprod[NUMPROBS][MAXSUPSIZE];
static double	optfilt[NUMPROBS][MAXSUPSIZE];
static double	filtsum[NUMPROBS];
static double	esqua[NUMPROBS];
static double	r_error[NUMPROBS];
static long	rhits[NUMPROBS];

static double	*miwork, *mfwork;

void reg_showit(int probno, int xsiz, int ysiz)
{
	sodump(optfilt[probno], xsiz, ysiz);
	printf("The post-filter error is %f\n", r_error[probno]);
}

int	Polymode, Qolymode;

void reg_init(int probno, int size)
{
	int	i, j;

	r_error[probno] = 0;
	esqua[probno] = rhits[probno] = 0;
	for (i = 0; i < size; i++)
	for (j = 0; j < size; j++)
		prods[probno][i][j] = 0;
	for (i = 0; i < size; i++)
		eprod[probno][i] = 0;
}

void rinit_all(int numprobs, int supsize)
{
	int	i;

	miwork = (double *)mustmalloc(supsize * supsize * sizeof(double), "Alloc miwork");
	mfwork = (double *)mustmalloc(supsize * supsize * sizeof(double), "Alloc mfwork");
	if (numprobs > NUMPROBS || supsize > MAXSUPSIZE) {
		printf("supsize too big\n");
		exit(EXIT_EXCEED);
	}
	for (i = 0; i < numprobs; i++)
		reg_init(i, supsize);
}

void upregress(double target, double *supp, int probno, int size)
{
	int	i, j;

	rhits[probno] += 1;
	esqua[probno] += target * target;
#if	1
	if (0) { /* } */
#else
	if (probno == 98) {
#endif
		printf("targ = %f S= ", target);
		for (i = 0; i < size; i++)
			printf("%f ", supp[i]);
		printf("\n");
		for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
			prods[probno][i][j] += supp[i] * supp[j];
		for (i = 0; i < size; i++)
			eprod[probno][i] += supp[i] * target;
	} else {
		for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
			prods[probno][i][j] += supp[i] * supp[j];
		for (i = 0; i < size; i++)
			eprod[probno][i] += supp[i] * target;
	}
}

void reg_answer(int probno, int size)
{
	int	i, j;

	submat(prods[probno][0], mfwork, MAXSUPSIZE, size, size);
	if (do_invert(mfwork, miwork, size) == 0) {
		/* Singular matrix */
		printf("singular matrix....\n");
		exit(EXIT_PEXCESS);
	}
	multiply(miwork, eprod[probno], optfilt[probno], size, size, 1);
	r_error[probno] = esqua[probno];
	for (i = 0; i < size; i++)
		for (j = 0; j < size; j++)
			r_error[probno] += optfilt[probno][i]
					* optfilt[probno][j]
					* prods[probno][i][j];
	filtsum[probno] = 0;
	for (i = 0; i < size; i++) {
		r_error[probno] -= 2 * optfilt[probno][i] * eprod[probno][i];
		filtsum[probno] += optfilt[probno][i];
	}
	if (rhits[probno]) {
		r_error[probno] /= rhits[probno];
	} else {
		r_error[probno] = 0;
	}
}

int main(int argc, char **argv)
{
	int	supsize, i;
	double	target, support[MAXSUPSIZE];

	if (argc == 3 && !strcmp(argv[1], "-poly")) {
		supsize = atoi(argv[2]) + 1;
		Polymode = 1;
	} else if (argc == 3 && !strcmp(argv[1], "-qoly")) {
		supsize = atoi(argv[2]) + 1;
		Qolymode = 1;
	} else if (argc > 2) {
	usage:
		fprintf(stderr, "Usage: regression [#M]\n");
		fprintf(stderr, " accepts (M+1)-tuples from stdin and\n");
		fprintf(stderr, " outputs the M-filter which produces target\n");
		fprintf(stderr, "If \"#M\" is omitted, M is first read from stdin\n");
		fprintf(stderr, "Usage: regression -poly #M\n");
		fprintf(stderr, " fits to N-degree poly of one input\n");
		fprintf(stderr, "Usage: regression -qoly #M\n");
		fprintf(stderr, " fits to N-1 degree poly of one input\n");
		fprintf(stderr, " with 1/input last\n");
		fprintf(stderr, "Remember to include an always-1 field in input tuple to get additive constant\n");
		exit(EXIT_USAGE);
	} else if (argc == 2) {
		supsize = atoi(argv[1]);
		if (supsize < 1)
			goto usage;
	} else
		scanf("%d", &supsize);
	rinit_all(NUMPROBS, supsize);
	support[0] = 1;
	while (1) {
		if (Polymode) {
			if (1 != scanf("%lf", support + 1))
				break;
			for (i = 2; i < supsize; i++)
				support[i] = support[i - 1] * support[1];
			support[0] = 1;
		} else if (Qolymode) {
			if (1 != scanf("%lf", support + 1))
				break;
			for (i = 2; i < supsize - 1; i++)
				support[i] = support[i - 1] * support[1];
			support[i] = 1 /  support[1];
			support[0] = 1;
		} else {
			for (i = 0; i < supsize; i++)
				if (1 != scanf("%lf", support + i))
					break;
			if (i < supsize)
				break;
		}
		if (1 != scanf("%lf", &target))
			break;
		upregress(target, support, 0, supsize);
	}
	reg_answer(0, supsize);
	if (supsize == 12)
		reg_showit(0, 2, 6);
	else
		reg_showit(0, supsize, 1);
	exit(EXIT_GOOD);
}

